/*
 * alarm_table.h
 *
 * copyleft: Sincrotrone Trieste S.C.p.A. di interesse nazionale
 *           Strada Statale 14 - km 163,5 in AREA Science Park
 *           34012 Basovizza, Trieste ITALY
 */

#ifndef ALARM_TABLE_H
#define ALARM_TABLE_H

#define _RW_LOCK

#include <iostream>
#include <string>
#include <map>
#include <numeric>		//for std::accumulate

#include <tango/tango.h>

//spirit defines have to be put befor first inclusion of spirit headers
#ifndef BOOST_SPIRIT_RULE_SCANNERTYPE_LIMIT
#define BOOST_SPIRIT_RULE_SCANNERTYPE_LIMIT 2 //tmp scanner_list
#endif

/*#ifndef BOOST_SPIRIT_THREADSAFE 
#define BOOST_SPIRIT_THREADSAFE 
#endif

#ifndef PHOENIX_THREADSAFE
#define PHOENIX_THREADSAFE
#endif
*/	
//#include "spirit-parser.h"
//#include <boost/spirit/core.hpp>
#include <boost/version.hpp>
#if BOOST_VERSION  < 103600
#include <boost/spirit/tree/ast.hpp>					//for ast parse trees (in tree_formula)
#else
#include <boost/spirit/include/classic_ast.hpp>			//for ast parse trees (in tree_formula)
#endif
//#define _USE_BOOST_LOCK
#ifdef _USE_BOOST_LOCK
#include <boost/thread/shared_mutex.hpp>
#endif

//#include "log_thread.h"

#define LOG_STREAM		TANGO_LOG

using namespace std;

typedef vector<Tango::DevDouble> value_t;

#if BOOST_VERSION  < 103600
typedef std::string::iterator  iterator_t;
typedef boost::spirit::node_val_data_factory<unsigned int> factory_t;		//want a modified node to contain an unsigned int value
typedef boost::spirit::tree_match<iterator_t, factory_t> parse_tree_match_t;     
typedef boost::spirit::tree_parse_info<iterator_t, factory_t>    tree_parse_info_t;   
#else
typedef std::string::iterator  iterator_t;
typedef boost::spirit::classic::node_val_data_factory<unsigned int> factory_t;		//want a modified node to contain an unsigned int value
typedef boost::spirit::classic::tree_match<iterator_t, factory_t> parse_tree_match_t;     
typedef boost::spirit::classic::tree_parse_info<iterator_t, factory_t>    tree_parse_info_t;   
#endif

typedef parse_tree_match_t::tree_iterator iter_t;

#define S_NORMAL	"NORMAL"
#define S_ALARM		"ALARM"
#define S_ERROR		"ERROR"

#define NOT_ACK		"NACK"
#define ACK		"ACK"

#define GR_ALL		0xffffffff
#define GR_NONE		0x00000000
#define GR_DEFAULT	GR_NONE			//or GR_ALL??
//#define MAX_GRP		32
#define GR_ALL_NAME		"gr_all"
#define GR_NONE_NAME	"gr_none"

#define LEV_LOG			"log"
#define LEV_WARNING		"warning"
#define LEV_FAULT		"fault"
//#define LEV_DEFAULT	LEV_FAULT

#define LEV_LOWEST		"lowest"
#define LEV_LOW			"low"
#define LEV_MEDIUM		"medium"
#define LEV_HIGH		"high"
#define LEV_HIGHEST		"highest"
#define LEV_DEFAULT		LEV_HIGH

class alarm_t;
class alarm_table;
class log_thread;
class cmd_thread;

#ifdef _USE_BOOST_LOCK
struct rwlock_t
{
	string name;
	rwlock_t(string n){name=n;};
	rwlock_t(){name=string("RWLOCK");};
	boost::shared_mutex mut;
	void readerIn(){TANGO_LOG<<name<<": " << __func__<<endl;mut.lock_shared();}
	void readerOut(){TANGO_LOG<<name<<": " << __func__<<endl;mut.unlock_shared();}
	void writerIn(){TANGO_LOG<<name<<": " << __func__<<endl;mut.lock();}
	void writerOut(){TANGO_LOG<<name<<": " << __func__<<endl;mut.unlock();}
};
#endif

template <typename T>
string print_vector(const std::vector<T>& v, const char * const separator = ",")
{
	ostringstream out;
    if(!v.empty())
    {
        std::copy(v.begin(),
                  --v.end(),
                  std::ostream_iterator<T>(out/*std::cout*/, separator));
        out << v.back();
    }
	return out.str();
}

struct vectorUtils
{
static double fBitAnd(long a, long b){	return (double)(a & b);}
static double fBitOr(long a, long b){	return (double)(a | b);}
static double fBitXor(long a, long b){	return (double)(a ^ b);}
static double fBitShiftL(long a, long b){	return (double)(a << b);}
static double fBitShiftR(long a, long b){	return (double)(a >> b);}
static double fMin(double a, double b){	return (double)min(a, b);}
static double fMax(double a, double b){	return (double)max(a, b);}
static double fPow(double a, double b){	return (double)pow(a, b);}
static double fEqual(double a, double b){	return (double)(a == b);}
static double fDifferent(double a, double b){	return (double)(a != b);}
static double fGreaterEqual(double a, double b){	return (double)(a >= b);}
static double fGreater(double a, double b){	return (double)(a > b);}
static double fLessEqual(double a, double b){	return (double)(a <= b);}
static double fLess(double a, double b){	return (double)(a < b);}
static double fPlus(double a, double b){	return (double)(a + b);}
static double fMinus(double a, double b){	return (double)(a - b);}
static double fMult(double a, double b){	return (double)(a * b);}
static double fDiv(double a, double b){	return (double)(a / b);}
static double fAnd(double a, double b){	return (double)((bool)a && (bool)b);}
static double fOr(double a, double b){	return (double)((bool)a || (bool)b);}

template<class T >
static void applyVectorFunc(const vector<T> & v1, long v1_x, long v1_y, const vector<T> & v2, long v2_x, long v2_y, value_t &res, long &res_x, long &res_y, const string& origin, double (*func)(T, T))
{
	if(v1.size() > 0 && v2.size() > 0 && (v1.size() == v2.size()))
	{
		res_x = v1_x;
		res_y = v1_y;
		if(v1_x != v2_x || (v1_y != v2_y && (v1_y > 1 || v2_y > 1)))//dim_y 0 and 1 are the same
		{
			ostringstream o;
			o << "Array sizes do not match: " << (v1_y ? v1_y : 1) << "x" << v1_x <<  " != " << (v2_y ? v2_y : 1) << "x" << v2_x;
			Tango::Except::throw_exception( //throw exception to have error not delayed
				"FORMULA_ERROR",
				o.str(),
				origin);
		}
		res.resize(v1.size());
		std::transform(v1.begin(), v1.end(), v2.begin(),res.begin(),
					[func](T a, T b) { return func(a, b); });				
	}
	else if(v2.size() == 1)
	{
		res_x = v1_x;
		res_y = v1_y;
		T b = v2[0];
		res.resize(v1.size());
		std::transform(v1.begin(), v1.end(), res.begin(),
					[b,func](T a) { return func(a, b); });
	}
	else if(v1.size() == 1)
	{
		res_x = v2_x;
		res_y = v2_y;
		T a = v1[0];
		res.resize(v2.size());
		std::transform(v2.begin(), v2.end(), res.begin(),
					[a,func](T b) { return func(a, b); });
	}
	else if(v1.size() > 0 && v2.size() > 0)
	{
		ostringstream o;
		o << "Array sizes do not match: " << (v1_y ? v1_y : 1) << "x" << v1_x <<  " != " << (v2_y ? v2_y : 1) << "x" << v2_x;
		Tango::Except::throw_exception( //throw exception to have error not delayed
			"FORMULA_ERROR",
			o.str(),
			origin);
	}
}
};

struct formula_res_t
{
	formula_res_t(){valid=false;quality=Tango::ATTR_VALID;ex_reason=string("");ex_desc=string("");ex_origin=string("");dim_x=0;dim_y=0;}
	value_t value;
	long dim_x;
	long dim_y;
	int quality;
	bool valid;
	string error;
	Tango::DevErrorList 	errors;	//TODO: error stack
	string ex_reason;
	string ex_desc;
	string ex_origin;
	string attr_name; //last evaluated attr name
	int combine_quality(int quality1, int quality2)
	{
		int res;
		if(quality1 == Tango::ATTR_INVALID ||  quality2 == Tango::ATTR_INVALID)
			res = Tango::ATTR_INVALID;
		else if(quality1 == Tango::ATTR_ALARM ||  quality2 == Tango::ATTR_ALARM)
			res = Tango::ATTR_ALARM;
		else if(quality1 == Tango::ATTR_WARNING ||  quality2 == Tango::ATTR_WARNING)
			res = Tango::ATTR_WARNING;
		else if(quality1 == Tango::ATTR_CHANGING ||  quality2 == Tango::ATTR_CHANGING)
			res = Tango::ATTR_CHANGING;
		else
			res = (quality1 > quality2) ? quality1 : quality2;	//TODO: decide priority in enum AttrQuality { ATTR_VALID, ATTR_INVALID, ATTR_ALARM, ATTR_CHANGING, ATTR_WARNING /*, __max_AttrQuality=0xffffffff */ };
		return res;
	}
	string combine_exception(string ex_1, string ex_2)
	{
		if(ex_1.length() > 0)
			return ex_1;
		else
			return ex_2;
	}
	void extract_result()
	{
		value = {(double)std::accumulate(value.begin(), value.end(), (bool)false, [](bool a, double b) { return (bool)(a || (bool)b); })};
	}

	formula_res_t operator==(const formula_res_t& e)
		{
			formula_res_t res;
			res.valid = valid && e.valid;
			res.quality = combine_quality(quality, e.quality);
			if(!res.valid)
			{
				res.ex_reason = combine_exception(ex_reason, e.ex_reason);
				res.ex_desc = combine_exception(ex_desc, e.ex_desc);
				res.ex_origin = combine_exception(ex_origin, e.ex_origin);
			}
			vectorUtils::applyVectorFunc(value,dim_x,dim_y,e.value,e.dim_x,e.dim_y,res.value,res.dim_x,res.dim_y,__func__,vectorUtils::fEqual);
			return res;
		}
	formula_res_t operator!=(const formula_res_t& e)
		{
			formula_res_t res;
			res.valid = valid && e.valid;
			res.quality = combine_quality(quality, e.quality);
			if(!res.valid)
			{
				res.ex_reason = combine_exception(ex_reason, e.ex_reason);
				res.ex_desc = combine_exception(ex_desc, e.ex_desc);
				res.ex_origin = combine_exception(ex_origin, e.ex_origin);
			}
			vectorUtils::applyVectorFunc(value,dim_x,dim_y,e.value,e.dim_x,e.dim_y,res.value,res.dim_x,res.dim_y,__func__,vectorUtils::fDifferent);
			return res;
		}
	formula_res_t operator<=(const formula_res_t& e)
		{
			formula_res_t res;
			res.valid = valid && e.valid;
			res.quality = combine_quality(quality, e.quality);
			if(!res.valid)
			{
				res.ex_reason = combine_exception(ex_reason, e.ex_reason);
				res.ex_desc = combine_exception(ex_desc, e.ex_desc);
				res.ex_origin = combine_exception(ex_origin, e.ex_origin);
			}
			vectorUtils::applyVectorFunc(value,dim_x,dim_y,e.value,e.dim_x,e.dim_y,res.value,res.dim_x,res.dim_y,__func__,vectorUtils::fLessEqual);
			return res;
		}
	formula_res_t operator>=(const formula_res_t& e)
		{
			formula_res_t res;
			res.valid = valid && e.valid;
			res.quality = combine_quality(quality, e.quality);
			if(!res.valid)
			{
				res.ex_reason = combine_exception(ex_reason, e.ex_reason);
				res.ex_desc = combine_exception(ex_desc, e.ex_desc);
				res.ex_origin = combine_exception(ex_origin, e.ex_origin);
			}
			vectorUtils::applyVectorFunc(value,dim_x,dim_y,e.value,e.dim_x,e.dim_y,res.value,res.dim_x,res.dim_y,__func__,vectorUtils::fGreaterEqual);
			return res;
		}
	formula_res_t operator<(const formula_res_t& e)
		{
			formula_res_t res;
			res.valid = valid && e.valid;
			res.quality = combine_quality(quality, e.quality);
			if(!res.valid)
			{
				res.ex_reason = combine_exception(ex_reason, e.ex_reason);
				res.ex_desc = combine_exception(ex_desc, e.ex_desc);
				res.ex_origin = combine_exception(ex_origin, e.ex_origin);
			}
			vectorUtils::applyVectorFunc(value,dim_x,dim_y,e.value,e.dim_x,e.dim_y,res.value,res.dim_x,res.dim_y,__func__,vectorUtils::fLess);
			return res;
		}
	formula_res_t operator>(const formula_res_t& e)
		{
			formula_res_t res;
			res.valid = valid && e.valid;
			res.quality = combine_quality(quality, e.quality);
			if(!res.valid)
			{
				res.ex_reason = combine_exception(ex_reason, e.ex_reason);
				res.ex_desc = combine_exception(ex_desc, e.ex_desc);
				res.ex_origin = combine_exception(ex_origin, e.ex_origin);
			}
			vectorUtils::applyVectorFunc(value,dim_x,dim_y,e.value,e.dim_x,e.dim_y,res.value,res.dim_x,res.dim_y,__func__,vectorUtils::fGreater);
			return res;
		}
	formula_res_t operator||(const formula_res_t& e)
		{
			formula_res_t res;
			if(value.size() <= 1 && e.value.size() <= 1) //apply special logic ov validity only if comparing scalars. Being careful that if valid=false, then size could be 0
			{
				res.value = {(double)((valid && value[0]) || (e.valid && e.value[0]))}; //in OR one operand TRUE and valid is enough
				res.valid = (valid && value[0]) || (e.valid && e.value[0]) || (valid && e.valid); //in OR one operand TRUE and valid is enough, otherwise both valid as usual 
			}
			else
				res.valid = valid && e.valid;
			res.quality = combine_quality(quality, e.quality);
			if(!res.valid)
			{
				res.ex_reason = combine_exception(ex_reason, e.ex_reason);
				res.ex_desc = combine_exception(ex_desc, e.ex_desc);
				res.ex_origin = combine_exception(ex_origin, e.ex_origin);
			}
			if(!(value.size() <= 1 && e.value.size() <= 1))
			{
				vectorUtils::applyVectorFunc(value,dim_x,dim_y,e.value,e.dim_x,e.dim_y,res.value,res.dim_x,res.dim_y,__func__,vectorUtils::fOr);
			}
			return res;
		}
	formula_res_t operator&&(const formula_res_t& e)
		{
			formula_res_t res;
			if(value.size() <= 1 && e.value.size() <= 1) //apply special logic ov validity only if comparing scalars
			{
				res.value = {(double)(!((valid && !value[0]) || (e.valid && !e.value[0])) || (valid && value[0] && e.valid && e.value[0]))}; //if at least one operand is FALSE and valid, result is FALSE for sure (and valid)
				res.valid = (valid && !value[0]) || (e.valid && !e.value[0]) || (valid && e.valid);
			}
			else
				res.valid = valid && e.valid;
			res.quality = combine_quality(quality, e.quality);
			if(!res.valid)
			{
				res.ex_reason = combine_exception(ex_reason, e.ex_reason);
				res.ex_desc = combine_exception(ex_desc, e.ex_desc);
				res.ex_origin = combine_exception(ex_origin, e.ex_origin);
			}
			if(!(value.size() <= 1 && e.value.size() <= 1))
			{
				vectorUtils::applyVectorFunc(value,dim_x,dim_y,e.value,e.dim_x,e.dim_y,res.value,res.dim_x,res.dim_y,__func__,vectorUtils::fAnd);
			}
			return res;
		}
	formula_res_t operator+(const formula_res_t& e)
		{
			formula_res_t res;
			res.valid = valid && e.valid;
			res.quality = combine_quality(quality, e.quality);
			if(!res.valid)
			{
				res.ex_reason = combine_exception(ex_reason, e.ex_reason);
				res.ex_desc = combine_exception(ex_desc, e.ex_desc);
				res.ex_origin = combine_exception(ex_origin, e.ex_origin);
			}
			vectorUtils::applyVectorFunc(value,dim_x,dim_y,e.value,e.dim_x,e.dim_y,res.value,res.dim_x,res.dim_y,__func__,vectorUtils::fPlus);
			return res;
		}
	formula_res_t operator-(const formula_res_t& e)
		{
			formula_res_t res;
			res.valid = valid && e.valid;
			res.quality = combine_quality(quality, e.quality);
			if(!res.valid)
			{
				res.ex_reason = combine_exception(ex_reason, e.ex_reason);
				res.ex_desc = combine_exception(ex_desc, e.ex_desc);
				res.ex_origin = combine_exception(ex_origin, e.ex_origin);
			}
			vectorUtils::applyVectorFunc(value,dim_x,dim_y,e.value,e.dim_x,e.dim_y,res.value,res.dim_x,res.dim_y,__func__,vectorUtils::fMinus);
			return res;
		}
	formula_res_t operator*(const formula_res_t& e)
		{
			formula_res_t res;
			res.valid = valid && e.valid;
			res.quality = combine_quality(quality, e.quality);
			if(!res.valid)
			{
				res.ex_reason = combine_exception(ex_reason, e.ex_reason);
				res.ex_desc = combine_exception(ex_desc, e.ex_desc);
				res.ex_origin = combine_exception(ex_origin, e.ex_origin);
			}
			vectorUtils::applyVectorFunc(value,dim_x,dim_y,e.value,e.dim_x,e.dim_y,res.value,res.dim_x,res.dim_y,__func__,vectorUtils::fMult);
			return res;
		}
	formula_res_t operator/(const formula_res_t& e)
		{
			formula_res_t res;
			res.valid = valid && e.valid;
			res.quality = combine_quality(quality, e.quality);
			if(!res.valid)
			{
				res.ex_reason = combine_exception(ex_reason, e.ex_reason);
				res.ex_desc = combine_exception(ex_desc, e.ex_desc);
				res.ex_origin = combine_exception(ex_origin, e.ex_origin);
			}
			vectorUtils::applyVectorFunc(value,dim_x,dim_y,e.value,e.dim_x,e.dim_y,res.value,res.dim_x,res.dim_y,__func__,vectorUtils::fDiv);
			return res;
		}
	formula_res_t operator!()
		{
			formula_res_t res;
			for(size_t ii=0; ii < value.size(); ii++)
			{
				res.value.push_back(!value[ii]);
			}
			return res;
		}
	formula_res_t operator-()
		{
			formula_res_t res;
			for(size_t ii=0; ii < value.size(); ii++)
			{
				res.value.push_back(-value[ii]);
			}
			return res;
		}
	/*string operator<<(const formula_res_t& e)
		{
			stringstream res;
			res << "value="<<e.value<<" quality="<<e.quality<<" EX reason="<<e.ex_reason<<" desc="<<e.ex_desc<<" origin="<<e.ex_origin;
			return res.str();
		}*/
};

/*
 * store the alarm-name/alarm-formula pair
 */
class alarm_t {
	public:
		string name,
					 formula;
		string attr_name;
		string attr_name_formula;
		Tango::DevEnum *attr_value;
		Tango::DevString *attr_value_formula;
		int quality;
		string ex_reason;
		string ex_desc;
		string ex_origin;
		Tango::TimeVal ts;
		string stat{S_NORMAL},
					 ack{ACK};
		bool error;
		bool enabled{true};
		bool shelved{false};
#ifdef _CNT_ATOMIC
		atomic_uint on_counter= {0};
		atomic_uint off_counter= {0};
		atomic_uint err_counter= {0};
#else
		unsigned int on_counter= {0};
		unsigned int off_counter= {0};
		unsigned int err_counter= {0};
#endif
		unsigned int freq_counter{0};
		
		tree_parse_info_t formula_tree;
					 
		static map<string, unsigned int> grp_str;

		bool done;
		bool to_be_evaluated;
		string msg;
		string url;
		string receivers;
		unsigned int grp{0};
		string lev;
		set<string> s_event;
		int is_new;
		Tango::TimeVal ts_on_delay;	//says when it has gone in alarm status for the first time
		unsigned int on_delay{0};		//TODO: seconds, is it enough precision?
		Tango::TimeVal ts_off_delay;	//says when it returned normal status
		unsigned int off_delay{0};		//TODO: seconds, is it enough precision?
		Tango::TimeVal ts_err_delay;	//says when it has gone in error status for the first time

		Tango::TimeVal ts_time_silenced;	//says when it has been silenced
		int silent_time{-1};			//minutes max to be silent
		int silenced;				//minutes still to be silent
		string attr_values;		//attr_values
		string attr_values_delay;	//attr_values of first occurrence of alarm waiting for on or off delay
		string cmd_name_a{""};					//action to execute: when NORMAL -> ALARM, cmd_name = cmd_dp_a/cmd_action_a
		string cmd_dp_a;						//device proxy part of cmd_name_a
		string cmd_action_a;					//action part of cmd_name_a
		bool send_arg_a;					//send as string argument alarm name and attr values
		Tango::DeviceProxy *dp_a;
		string cmd_name_n{""};					//action to execute: when ALARM -> NORMAL, cmd_name_n = cmd_dp_n/cmd_action_n
		string cmd_dp_n;						//device proxy part of cmd_name_n
		string cmd_action_n;					//action part of cmd_name_n
		bool send_arg_n;					//send as string argument alarm name and attr values
		Tango::DeviceProxy *dp_n;
		/*
		 * methods
		 */
		alarm_t() noexcept {};							//constructor
#ifdef _CNT_ATOMIC
		alarm_t& operator=(const alarm_t& rhs) { on_counter = rhs.on_counter.load(); off_counter = rhs.off_counter.load(); err_counter = rhs.err_counter.load(); return *this; }
		alarm_t(const alarm_t& rhs) { on_counter = rhs.on_counter.load(); off_counter = rhs.off_counter.load(); err_counter = rhs.err_counter.load();}
#endif
		void init_static_map(vector<string> &group_names);
		bool operator==(const alarm_t& that);
		bool operator==(const string& n);
		void str2alm(const string &s);
		string alm2str(void);
		string grp2str(void);
		void add_grp_from_str(string &s);
		void str2grp(string &s);
		void insert(const string& s);
		void clear();
		void confstr(string &s);
		
	protected:
	private:
		
};

typedef map<string,alarm_t> alarm_container_t;	
class alarm_table {
	public:
		alarm_table() {}
		~alarm_table() {del_rwlock();}
		bool setAlarmQuality {false};
		void set_dev(Tango::DeviceImpl* devImpl) {mydev=devImpl;}
		void set_al_qual(bool set_al_qual) {setAlarmQuality=set_al_qual;}

		//void init(vector<string>& avs);
		//void init(vector<string>& avs, vector<string> &evn, map< string,vector<string> > &alarm_event);		
		void push_back(alarm_t& a);
		void show(vector<string> &al_table_string);
		unsigned int size(void);
		alarm_container_t& get(void);
		void stored(vector<alarm_t>& a);
		bool update(const string& alm_name, Tango::TimeVal ts, formula_res_t res, string &attr_values, string grp, string msg, string formula);
		bool timer_update();
		void erase(alarm_container_t::iterator i);
		bool exist(string& s);
		unsigned int to_be_evaluated_num();
		vector<string> to_be_evaluated_list();
		//vector<alarm_t> v_alarm;
		alarm_container_t v_alarm;
#ifndef _USE_BOOST_LOCK
		ReadersWritersLock *vlock;
#else
		rwlock_t *vlock;
#endif
		void new_rwlock();
		void del_rwlock();

		void save_alarm_conf_db(const string &att_name, const string &name, const string &status, const string &ack, bool enabled,
				 const string &formula, unsigned int on_delay, unsigned int off_delay, const string &grp, const string &url, const string &lev, const string &msg, const string &cmd_a, const string &cmd_n, const string &receivers, int silent_time);
		void delete_alarm_conf_db(string att_name);
		void get_alarm_list_db(vector<string> &al_list, map<string, string> &saved_alarms, ReadersWritersLock *savedlock);
		void init_cmdthread();
		void stop_cmdthread();
		Tango::TimeVal startup_complete;			//to disable action execution at startup
		void set_err_delay(unsigned int delay){err_delay=delay;};
	
	protected:
	private:

		Tango::DeviceImpl* mydev;
		log_thread *logloop;
		cmd_thread *cmdloop;
		unsigned int err_delay;
};


#endif	/* ALARM_TABLE_H */
